<?php

namespace IQBall\Api\Controller;

use IQBall\Api\APIControl;
use IQBall\Core\Data\Account;
use IQBall\Core\Data\TacticInfo;
use IQBall\Core\Http\HttpCodes;
use IQBall\Core\Http\HttpRequest;
use IQBall\Core\Http\HttpResponse;
use IQBall\Core\Http\JsonHttpResponse;
use IQBall\Core\Model\TacticModel;
use IQBall\Core\Validation\DefaultValidators;

/**
 * API endpoint related to tactics
 */
class APITacticController {
    private TacticModel $model;

    /**
     * @param TacticModel $model
     */
    public function __construct(TacticModel $model) {
        $this->model = $model;
    }

    /**
     * update name of tactic, specified by tactic identifier, given in url.
     * @param int $tactic_id
     * @param Account $account
     * @return HttpResponse
     */
    public function updateName(int $tactic_id, Account $account): HttpResponse {
        return APIControl::runChecked([
            "name" => [DefaultValidators::lenBetween(1, 50), DefaultValidators::nameWithSpaces()],
        ], function (HttpRequest $request) use ($tactic_id, $account) {

            $failures = $this->model->updateName($tactic_id, $request["name"], $account->getUser()->getId());

            if (!empty($failures)) {
                //TODO find a system to handle Unauthorized error codes more easily from failures.
                return new JsonHttpResponse($failures, HttpCodes::BAD_REQUEST);
            }

            return HttpResponse::fromCode(HttpCodes::OK);
        });
    }

    /**
     * @param int $id
     * @param Account $account
     * @return HttpResponse
     */
    public function saveContent(int $id, Account $account): HttpResponse {
        return APIControl::runChecked([
            "content" => [],
        ], function (HttpRequest $req) use ($id) {
            //TODO verify that the account has the rights to update the tactic content
            if ($fail = $this->model->updateContent($id, json_encode($req["content"]))) {
                return new JsonHttpResponse([$fail], HttpCodes::BAD_REQUEST);
            }
            return HttpResponse::fromCode(HttpCodes::OK);
        });
    }


    /**
     * @param int $userId
     * @return HttpResponse given user information.
     */
    public function getUserTactics(int $userId): HttpResponse {
        $tactics = $this->model->listAllOf($userId);

        $response = array_map(fn(TacticInfo $t) => [
            'id' => $t->getId(),
            'name' => $t->getName(),
            'court' => $t->getCourtType(),
            'creation_date' => $t->getCreationDate(),

        ], $tactics);

        return new JsonHttpResponse($response);
    }


}
