<?php

namespace IQBall\Core;

use IQBall\Core\Http\HttpCodes;
use IQBall\Core\Http\HttpRequest;
use IQBall\Core\Http\HttpResponse;
use IQBall\Core\Http\JsonHttpResponse;
use IQBall\Core\Validation\ValidationFail;
use IQBall\Core\Validation\Validator;

class Control {
    /**
     * Runs given callback, if the request's payload json validates the given schema.
     * @param array<string, Validator[]> $schema an array of `fieldName => DefaultValidators` which represents the request object schema
     * @param callable(HttpRequest): HttpResponse $run the callback to run if the request is valid according to the given schema.
     *                      The callback must accept an HttpRequest, and return an HttpResponse object.
     * @param ControlSchemaErrorResponseFactory $errorFactory an error factory to use if the request does not validate the required schema
     * @return HttpResponse
     */
    public static function runChecked(array $schema, callable $run, ControlSchemaErrorResponseFactory $errorFactory): HttpResponse {
        $request_body = file_get_contents('php://input');
        $payload_obj = json_decode($request_body);
        if (!$payload_obj instanceof \stdClass) {
            $fail = new ValidationFail("bad-payload", "request body is not a valid json object");
            return $errorFactory->apply([$fail]);

        }
        $payload = get_object_vars($payload_obj);
        return self::runCheckedFrom($payload, $schema, $run, $errorFactory);
    }

    /**
     * Runs given callback, if the given request data array validates the given schema.
     * @param array<string, mixed> $data the request's data array.
     * @param array<string, Validator[]> $schema an array of `fieldName => DefaultValidators` which represents the request object schema
     * @param callable(HttpRequest): HttpResponse $run the callback to run if the request is valid according to the given schema.
     *                      The callback must accept an HttpRequest, and return an HttpResponse object.
     * @param ControlSchemaErrorResponseFactory $errorFactory an error factory to use if the request does not validate the required schema
     * @return HttpResponse
     */
    public static function runCheckedFrom(array $data, array $schema, callable $run, ControlSchemaErrorResponseFactory $errorFactory): HttpResponse {
        $fails = [];
        $request = HttpRequest::from($data, $fails, $schema);

        if (!empty($fails)) {
            return $errorFactory->apply($fails);

        }

        return call_user_func_array($run, [$request]);
    }
}
