<?php

namespace IQBall\Core\Gateway;

use IQBall\Core\Connection;
use IQBall\Core\Data\Account;
use IQBall\Core\Data\User;
use PDO;

class AccountGateway {
    private Connection $con;

    /**
     * @param Connection $con
     */
    public function __construct(Connection $con) {
        $this->con = $con;
    }

    public function insertAccount(string $name, string $email, string $token, string $hash, string $profilePicture): int {
        $this->con->exec("INSERT INTO Account(username, hash, email, token,profile_picture) VALUES (:username,:hash,:email,:token,:profile_pic)", [
            ':username' => [$name, PDO::PARAM_STR],
            ':hash' => [$hash, PDO::PARAM_STR],
            ':email' => [$email, PDO::PARAM_STR],
            ':token' => [$token, PDO::PARAM_STR],
            ':profile_pic' => [$profilePicture, PDO::PARAM_STR],
        ]);
        return intval($this->con->lastInsertId());
    }

    public function updateAccount(int $id, string $name, string $email, string $token, bool $isAdmin): void {
        $this->con->exec("UPDATE Account SET username = :username, email = :email, token = :token WHERE id = :id", [
            ':username' => [$name, PDO::PARAM_STR],
            ':email' => [$email, PDO::PARAM_STR],
            ':token' => [$token, PDO::PARAM_STR],
            ':id' => [$id, PDO::PARAM_INT],
        ]);
        $this->setIsAdmin($id, $isAdmin);
    }

    public function isAdmin(int $id): bool {
        $stmnt = $this->con->prepare("SELECT * FROM Admins WHERE id = :id");
        $stmnt->bindValue(':id', $id, PDO::PARAM_INT);
        $stmnt->execute();
        $result = $stmnt->fetchAll(PDO::FETCH_ASSOC);

        return !empty($result);
    }

    /**
     * promote or demote a user to server administrator
     * @param int $id
     * @param bool $isAdmin true to promote, false to demote
     * @return bool true if the given user exists
     */
    public function setIsAdmin(int $id, bool $isAdmin): bool {
        if ($isAdmin) {
            $stmnt = $this->con->prepare("INSERT INTO Admins VALUES(:id)");
        } else {
            $stmnt = $this->con->prepare("DELETE FROM Admins WHERE id = :id");
        }

        $stmnt->bindValue(':id', $id);
        $stmnt->execute();

        return $stmnt->rowCount() > 0;
    }

    /**
     * @param string $email
     * @return array<string, mixed>|null
     */
    private function getRowsFromMail(string $email): ?array {
        return $this->con->fetch("SELECT * FROM Account WHERE email = :email", [':email' => [$email, PDO::PARAM_STR]])[0] ?? null;
    }

    /**
     * @param string $email
     * @return string|null the hashed user's password, or null if the given mail does not exist
     */
    public function getHash(string $email): ?string {
        $results = $this->getRowsFromMail($email);
        if ($results == null) {
            return null;
        }
        return $results['hash'];
    }

    /**
     * @param string $email
     * @return bool true if the given email exists in the database
     */
    public function exists(string $email): bool {
        return $this->getRowsFromMail($email) != null;
    }

    /**
     * @param string $email
     * @return Account|null
     */
    public function getAccountFromMail(string $email): ?Account {
        $acc = $this->getRowsFromMail($email);
        if (empty($acc)) {
            return null;
        }

        return new Account($acc["token"], new User($email, $acc["username"], $acc["id"], $acc["profile_picture"], $this->isAdmin($acc["id"])));
    }

    /**
     * @param string $token get an account from given token
     * @return Account|null
     */
    public function getAccountFromToken(string $token): ?Account {
        $stmnt = $this->con->prepare("SELECT * FROM Account WHERE token = :token");
        $stmnt->bindValue(':token', $token);
        return $this->getAccountFrom($stmnt);
    }

    /**
     * @param int $id get an account from given identifier
     * @return Account|null
     */
    public function getAccount(int $id): ?Account {
        $stmnt = $this->con->prepare("SELECT * FROM Account WHERE id = :id");
        $stmnt->bindValue(':id', $id);
        return $this->getAccountFrom($stmnt);
    }

    private function getAccountFrom(\PDOStatement $stmnt): ?Account {
        $stmnt->execute();
        $acc = $stmnt->fetch(PDO::FETCH_ASSOC);

        if ($acc == null) {
            return null;
        }

        return new Account($acc["token"], new User($acc["email"], $acc["username"], $acc["id"], $acc["profile_picture"], $this->isAdmin($acc["id"])));
    }

    /**
     * Return a list containing n accounts from a given starting index
     *
     * @param integer $n the number of accounts to retrieve
     * @param int $start starting index of the list content
     * @return Account[]
     */
    public function searchAccounts(int $start, int $n, ?string $searchString): array {
        $res = $this->con->fetch(
            "SELECT * FROM Account WHERE username LIKE '%' || :search || '%' OR email LIKE '%' || :search || '%' ORDER BY username, email LIMIT :offset, :n",
            [
                ":offset" => [$start, PDO::PARAM_INT],
                ":n" => [$n, PDO::PARAM_INT],
                ":search" => [$searchString ?? "", PDO::PARAM_STR],
            ]
        );
        return array_map(fn(array $acc) => new Account($acc["token"], new User($acc["email"], $acc["username"], $acc["id"], $acc["profile_picture"], $this->isAdmin($acc["id"]))), $res);

    }

    /**
     * returns the total amount of accounts in the database
     * @return int
     */
    public function totalCount(): int {
        return $this->con->fetch("SELECT count(*) FROM Account", [])[0]['count(*)'];
    }

    /**
     * remove a bunch of account identifiers
     * @param int[] $accountIds
     */
    public function removeAccounts(array $accountIds): void {
        foreach ($accountIds as $accountId) {
            $this->con->fetch("DELETE FROM Account WHERE id = :accountId", [
                ":accountId" => [$accountId, PDO::PARAM_INT],
            ]);
        }
    }
}
